<h4 class="report-title">General Summary</h4>
<div class="row mt-3 mb-5">
    <div class="col-6">
        <ul class="list-group">
            <li
                class="list-group-item d-flex justify-content-between align-items-center"
            >
                Processed Components
                <span class="badge bg-primary rounded-pill">{{
                    report_count
                }}</span>
            </li>
            {% if report_count - error_count > 0 %}
            <li
                class="list-group-item list-group-item-success d-flex justify-content-between align-items-center"
            >
                Successful Reports
                <span class="badge bg-primary rounded-pill">{{
                    report_count - error_count
                }}</span>
            </li>
            {% endif %} {% if error_count > 0 %}
            <li
                class="list-group-item list-group-item-danger d-flex justify-content-between align-items-center"
            >
                Failed Reports
                <span class="badge bg-primary rounded-pill">{{
                    error_count
                }}</span>
            </li>
            {% endif %}
        </ul>
    </div>
</div>
